/****************************************************************************
File    : input_win.c
/*
@(#) #SY# Atari800Win
@(#) #IS# Windows input implementation
@(#) #BY# Richard Lawrence, Harry (modifications and code maintenance)
@(#) #LM# 13.03.2000
*/

/*
Copyright (c) 1998 Richard Lawrence

This program is free software; you can redistribute it and/or modify it under the terms 
of the GNU General Public License as published by the Free Software Foundation; either 
version 2 of the License, or (at your option) any later version. This program is 
distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. You should have received a copy of the GNU
General Public License along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <stdio.h>
#include <windows.h>
#include <windowsx.h>
#include <dinput.h>
#include "WinAtari.h"
#include "WinConfig.h"
#include "Resource.h"
#include "atari800.h"
#include "globals.h"
#include "display_win.h"
#include "misc_win.h"
#include "sound_win.h"
#include "registry.h"


/* Public objects */

HINSTANCE g_hDInput     = NULL;  /* DirectInput 5.0 interface object */
int       g_nCurrentKey = AKEY_NONE;
int       g_nCurrentVK  = 0;
int       g_nNewKey     = AKEY_NONE;
int       g_nNewVK      = 0;
int       g_nNumDIDevs  = 0;
int       CTRL_KEY;

char g_acInputDevNames      [ MAX_INPUT_DEVICES  ][ INPUT_DEV_NAMELEN ];
int  g_anSelectedInputDevice[ MAX_ATARI_JOYPORTS ] =
{
	-1, 127, 127, 127
};

/* Private objects */

static LPDIRECTINPUT        s_lpDInput = NULL;
static LPDIRECTINPUTDEVICE2 s_alpDevFound[ MAX_INPUT_DEVICES ];
static LPDIRECTINPUTDEVICE2 s_alpJoyPorts[ MAX_ATARI_JOYPORTS ] =
{
	NULL, NULL, NULL, NULL
};
static int  s_anTrig [ MAX_ATARI_JOYPORTS ];
static int  s_anStick[ MAX_ATARI_JOYPORTS ];

static int	s_nJoyCount   = 0;
static int  s_nConsol;
static BOOL	s_bBtn2Flip;
static BOOL	s_bBtn3Flip;
static BOOL	s_bBtn4Flip;
static BOOL s_bKBJoystick = FALSE;
static UINT s_unKeystick  = STICK_CENTRE;
static UINT s_unKeytrig   = 1;

static const int s_anStickTable[] =
{
	STICK_LL,
	STICK_BACK,
	STICK_LR,
	STICK_LEFT,
	STICK_CENTRE,
	STICK_RIGHT,
	STICK_UL,
	STICK_FORWARD,
	STICK_UR
};

static BOOL GetDIErrorString( HRESULT hResult, LPSTR lpszErrorBuff, DWORD cchError );
		

/*========================================================
Function : ResetKeys
=========================================================*/
/* #FN#
   Resets the keys state to default */
void
/* #AS#
   Nothing */
ResetKeys( void )
{
	s_unKeystick  = STICK_CENTRE;
	s_unKeytrig   = 1;
	g_nCurrentKey = g_nNewKey  = AKEY_NONE;
	g_nCurrentVK  = g_nNewVK   = 0;
	SHIFT_KEY     = CTRL_KEY   = FALSE;
	KEYPRESSED    = FALSE;
	s_nConsol     = 0x07;
} /* #OF# ResetKeys */

/*========================================================
Function : ResetInput
=========================================================*/
/* #FN#
   Resets the input (keys and joys) state to default */
void
/* #AS#
   Nothing */
ResetInput( void )
{
	int i;

	s_bBtn2Flip = FALSE;
	s_bBtn3Flip = FALSE;
	s_bBtn4Flip = FALSE;

	ResetKeys();
	for( i = 0; i < MAX_ATARI_JOYPORTS; i++ )
	{
		s_anTrig [ i ] = 1;
		s_anStick[ i ] = STICK_CENTRE;
	}
} /* #OF# ResetInput */

/*========================================================
Function : ShowDInputError
=========================================================*/
/* #FN#
   Fatal DirectInput error function. Since we cannot continue, bring us back
   to a known working state before displaying the error, then start an exit.
   Make sure to shut down the Atari layer on the way. */
static
void
/* #AS#
   Nothing */
ShowDInputError( UINT    nUID,    /* #IN# */
				 HRESULT hResult, /* #IN# */
				 BOOL    bQuit    /* #IN# */ )
{
	TCHAR tszError    [ LOADSTRING_STRING_SIZE ];
	TCHAR szFullString[ LOADSTRING_STRING_SIZE ];
	TCHAR szAction    [ LOADSTRING_STRING_SIZE ];
	
	/* Get us back to a GDI display and stop making noises */
	SetSafeDisplay( FALSE );
	ClearSound( FALSE );

	/* Get the error string and present it to the user */
	GetDIErrorString( hResult, tszError, LOADSTRING_STRING_SIZE );
	LoadString( NULL, IDS_DINPUT_ERROR_PROMPT, g_tszErrorString, LOADSTRING_STRING_SIZE );
	LoadString( NULL, nUID, szAction, LOADSTRING_STRING_SIZE );
	wsprintf( szFullString, g_tszErrorString, szAction, tszError );
	LoadString( NULL, IDS_DINPUT_ERROR_HDR, g_tszErrorString, LOADSTRING_STRING_SIZE );
	MessageBox( g_hMainWnd, szFullString, g_tszErrorString, MB_ICONSTOP | MB_OK );

	/* Start a quit (this will end up in Atari_Exit()) */
	if( bQuit )
	{
		/* Make sure the atari is turned off */
		g_ulAtariState = ATARI_UNINITIALIZED | ATARI_PAUSED;
		PostMessage( g_hMainWnd, WM_CLOSE, 0, 0L );
	}
	else
		RestartSound();
} /* #OF# ShowDInputError */

/*========================================================
Function : SetDIDwordProperty
=========================================================*/
/* #FN#
   Helper function just makes sure the Property struct is set correctly */
static
HRESULT
/* #AS#
   DI_OK, DI_PROPNOEFFECT, if the methods succeeds. DIERR_OBJECTNOTFOUND,
   DIERR_UNSUPPORTED, DIERR_NOTINITIALIZED, DIERR_INVALIDPARAM, if the method fails.
*/
SetDIDwordProperty( LPDIRECTINPUTDEVICE lpInputDev,   /* #IN# */
					REFGUID             guidProperty, /* #IN# */
					DWORD               dwObject,     /* #IN# */
					DWORD               dwHow,        /* #IN# */
					DWORD               dwValue       /* #IN# */ )
{
	DIPROPDWORD dipdw;
	
	dipdw.diph.dwSize       = sizeof(dipdw);
	dipdw.diph.dwHeaderSize = sizeof(dipdw.diph);
	dipdw.diph.dwObj        = dwObject;
	dipdw.diph.dwHow        = dwHow;
	dipdw.dwData            = dwValue;
	
	return lpInputDev->lpVtbl->SetProperty( lpInputDev, guidProperty, &dipdw.diph );
} /* #OF# SetDIDwordProperty */

static
BOOL FAR PASCAL
InitJoystickInput( LPCDIDEVICEINSTANCE lpDevInstance,
				   LPVOID pvRef )
{
	LPDIRECTINPUT       lpDirectInput = pvRef;
	LPDIRECTINPUTDEVICE pdev;
	DIPROPRANGE         diprg;
	HRESULT             hResult;
	
	/* Create the DirectInput joystick device */
	hResult = lpDirectInput->lpVtbl->CreateDevice( lpDirectInput, &lpDevInstance->guidInstance, &pdev, NULL);
	if( FAILED( hResult ) )
	{
		ShowDInputError( IDS_DIERR_CREATING_DEVICE, hResult, FALSE );
		return DIENUM_STOP;/*DIENUM_CONTINUE;*/
	}
	
	/* Set joystick data format */
	hResult = pdev->lpVtbl->SetDataFormat(pdev, &c_dfDIJoystick);
	if( FAILED( hResult ) )
	{
		ShowDInputError( IDS_DIERR_SETTING_JOYSTICK_FORMAT, hResult, FALSE );
		pdev->lpVtbl->Release(pdev);
		return DIENUM_STOP;
	}
	
	/* Set the cooperative level */
	hResult = pdev->lpVtbl->SetCooperativeLevel( pdev, g_hMainWnd, DISCL_NONEXCLUSIVE | DISCL_BACKGROUND );
	if( FAILED(hResult) )
	{
		ShowDInputError( IDS_DIERR_SETTING_COOP, hResult, FALSE );
		pdev->lpVtbl->Release(pdev);
		return DIENUM_STOP;
	}
	
	// set X-axis range to (-1000 ... +1000)
	// This lets us test against 0 to see which way the s_anStick is pointed.
	
	diprg.diph.dwSize       = sizeof(diprg);
	diprg.diph.dwHeaderSize = sizeof(diprg.diph);
	diprg.diph.dwObj        = DIJOFS_X;
	diprg.diph.dwHow        = DIPH_BYOFFSET;
	diprg.lMin              = -1000;
	diprg.lMax              = +1000;
	
	hResult = pdev->lpVtbl->SetProperty(pdev, DIPROP_RANGE, &diprg.diph);
	if( FAILED(hResult) )
	{
		ShowDInputError( IDS_DIERR_SETTING_PROPS, hResult, FALSE );
		pdev->lpVtbl->Release( pdev );
		return FALSE;
	}
	
	//
	// And again for Y-axis range
	//
	diprg.diph.dwObj        = DIJOFS_Y;
	
	hResult = pdev->lpVtbl->SetProperty(pdev, DIPROP_RANGE, &diprg.diph);
	if( FAILED( hResult ) )
	{
		ShowDInputError( IDS_DIERR_SETTING_PROPS, hResult, FALSE );
		pdev->lpVtbl->Release( pdev );
		return FALSE;
	}
	
	// set X axis dead zone to 50% (to avoid accidental turning)
	// Units are ten thousandths, so 50% = 5000/10000.
	hResult = SetDIDwordProperty(pdev, DIPROP_DEADZONE, DIJOFS_X, DIPH_BYOFFSET, 5000);
	if( FAILED(hResult) )
	{
		ShowDInputError( IDS_DIERR_SETTING_DEADZONE, hResult, FALSE );
		pdev->lpVtbl->Release( pdev );
		return FALSE;
	}
	
	
	// set Y axis dead zone to 50% (to avoid accidental thrust)
	// Units are ten thousandths, so 50% = 5000/10000.
	hResult = SetDIDwordProperty(pdev, DIPROP_DEADZONE, DIJOFS_Y, DIPH_BYOFFSET, 5000);
	if( FAILED(hResult) )
	{
		ShowDInputError( IDS_DIERR_SETTING_DEADZONE, hResult, FALSE );
		pdev->lpVtbl->Release( pdev );
		return FALSE;
	}
	
	
	//
	// Add it to our list of devices.
	// 
	//AddInputDevice(pdev, pdinst);
	hResult = pdev->lpVtbl->QueryInterface( pdev, &IID_IDirectInputDevice2, (LPVOID *)&s_alpDevFound[g_nNumDIDevs]);
	if( FAILED(hResult) )
	{
		ShowDInputError( IDS_DIERR_ADDING_DEVICE, hResult, FALSE );
		return FALSE;
	}
	strncpy( g_acInputDevNames[ g_nNumDIDevs++ ], lpDevInstance->tszInstanceName, 32 );
	pdev->lpVtbl->Release( pdev );

	return DIENUM_CONTINUE;;
}

static BOOL ReacquireInput( int sticknum )
{
    HRESULT hResult;
	
    // if we have a current device
    if(s_alpJoyPorts[sticknum] )
    {
		// acquire the device
		hResult = IDirectInputDevice_Acquire(s_alpJoyPorts[sticknum]);
		if(SUCCEEDED(hResult))	// acquisition successful
			return TRUE;
		else					// acquisition failed
			return FALSE;
    }
    else	// we don't have a current device
		return FALSE;
}

void PickJoystick( int sticknum, int joystick )
{
	int i;

	for( i=0; i < MAX_ATARI_JOYPORTS; i++ )
	{
		if( s_alpJoyPorts[i] != NULL && s_alpJoyPorts[i] == s_alpDevFound[joystick] )
		{
			IDirectInputDevice_Unacquire(s_alpJoyPorts[i]);
			s_alpJoyPorts[i] = NULL;
		}
	}

	g_anSelectedInputDevice[sticknum] = joystick;

	if( s_alpJoyPorts[sticknum] )
		IDirectInputDevice_Unacquire(s_alpJoyPorts[sticknum]);
	s_alpJoyPorts[sticknum] = NULL;

	s_bKBJoystick = FALSE;
	for( i = 0; i < MAX_ATARI_JOYPORTS; i++ )
	{
		if( g_anSelectedInputDevice[i] < 0 )
			s_bKBJoystick = TRUE;
	}

	/* -1 means to use the keyboard instead */
	if( joystick < 0 )
		return;

	if( (joystick < MAX_INPUT_DEVICES) && s_alpDevFound[joystick] )
	{
		s_alpJoyPorts[ sticknum ] = s_alpDevFound[ joystick ];
		if( !ReacquireInput( sticknum ) )
			ShowDInputError( IDS_DIERR_ACQUIRE_INIT, DIERR_NOTACQUIRED, FALSE );
	}
}

int InitialiseInput( void )
{
	HRESULT	hResult;
	int s_anStick;
	char value;

	g_nCurrentKey = g_nNewKey = AKEY_NONE;
	g_nCurrentVK = g_nNewVK = 0;
	KEYPRESSED = FALSE;

	ResetInput();

	if( !s_lpDInput )
	{
		ZeroMemory( s_alpDevFound, sizeof( LPDIRECTINPUTDEVICE ) * MAX_INPUT_DEVICES );

		/* create the DirectInput 5.0 interface object */
		hResult = DirectInputCreate( g_hDInput, DIRECTINPUT_VERSION, &s_lpDInput, NULL );
		if( FAILED( hResult ) )
		{
			if( hResult == DIERR_OLDDIRECTINPUTVERSION )
			{
				s_lpDInput = NULL;
			}
			else
			{
				ShowDInputError( IDS_DIERR_CREATING_OBJ, hResult, FALSE );
				return 0;
			}
		}
		else
		{
			/* Enumerate the joystick devices. */
			
			hResult = s_lpDInput->lpVtbl->EnumDevices(s_lpDInput, DIDEVTYPE_JOYSTICK, InitJoystickInput, s_lpDInput, DIEDFL_ATTACHEDONLY);
			if( FAILED(hResult) )
			{
				ShowDInputError( IDS_DIERR_ENUMERATING, hResult, FALSE );
				return 0;
			}
			s_lpDInput->lpVtbl->Release(s_lpDInput);
		}
	}

	value = (char)(g_ulJoystickSelects & 0x000000ff);
	s_anStick = value;
	PickJoystick( 0, s_anStick );

	s_anStick = g_ulJoystickSelects & 0x0000ff00;
	s_anStick >>= 8;
	value = (char)s_anStick;
	s_anStick = value;
	if( s_anStick == g_anSelectedInputDevice[0] )
		s_anStick = NO_JOYSTICK;
	PickJoystick( 1, s_anStick );

	s_anStick = g_ulJoystickSelects & 0x00ff0000;
	s_anStick >>= 16;
	value = (char)s_anStick;
	s_anStick = value;
	if( s_anStick == g_anSelectedInputDevice[0] || s_anStick == g_anSelectedInputDevice[1] )
		s_anStick = NO_JOYSTICK;
	PickJoystick( 2, s_anStick );

	s_anStick = g_ulJoystickSelects & 0xff000000;
	s_anStick >>= 24;
	value = (char)s_anStick;
	s_anStick = value;
	if( s_anStick == g_anSelectedInputDevice[0] || s_anStick == g_anSelectedInputDevice[1] || s_anStick == g_anSelectedInputDevice[2] )
		s_anStick = NO_JOYSTICK;
	PickJoystick( 3, s_anStick );

	s_nJoyCount = g_nJoyRepeat;
	return 1;
}

void ClearInput(void)
{
	int idev;
	
	// make sure the device is unacquired
	// it doesn't harm to unacquire a device that isn't acquired
	for( idev = 0; idev < MAX_ATARI_JOYPORTS; idev++ )
	{
		if (s_alpJoyPorts[idev])
		{
			IDirectInputDevice_Unacquire(s_alpJoyPorts[idev]);
			s_alpJoyPorts[idev] = NULL;
		}
	}
	
	// release all the devices we created
	for (idev = 0; idev < g_nNumDIDevs; idev++)
	{
		if (s_alpDevFound[idev]) {
			IDirectInputDevice_Release(s_alpDevFound[idev]);
			s_alpDevFound[idev] = NULL;
		}
	}
}

int ReadJoystickInput( int sticknum, BOOL bTestButton )
{
	int						stickval = 0;
	HRESULT                 hResult;
	DIJOYSTATE              js;

	if( (sticknum >= MAX_INPUT_DEVICES) || !s_alpJoyPorts[sticknum] )
		return STICK_CENTRE;

	// poll the joystick to read the current state
	hResult = IDirectInputDevice2_Poll( s_alpJoyPorts[sticknum] );

	// get data from the joystick
	hResult = IDirectInputDevice_GetDeviceState( s_alpJoyPorts[sticknum], sizeof(DIJOYSTATE), &js );
	if( FAILED( hResult ) )
	{
		// did the read fail because we lost input for some reason?
		// if so, then attempt to reacquire.  If the second acquire
		// fails, then the error from GetDeviceData will be
		// DIERR_NOTACQUIRED, so we won't get stuck an infinite loop.
		if( hResult == DIERR_INPUTLOST )
		{
			ReacquireInput( sticknum );
			// return the fact that we did not read any data
			return 0;
		}
		ShowDInputError( IDS_DIERR_ACQUIRE, DIERR_NOTACQUIRED, FALSE );
	}

	if (machine == Atari5200) 
	{
		int i;
		for( i = 0; i < 4; i++ )
		{
			if( js.rgbButtons[i] & 0x80 )
				s_anTrig[i] = 0;
			else
				s_anTrig[i] = 1;
		}
	}
	else
	{
		if( !s_bBtn2Flip )
		{
			if( js.rgbButtons[1] & 0x80 )
			{
				Toggle_FullSpeed();
				s_bBtn2Flip = TRUE;
			}
		}
		else
		{
			if( !(js.rgbButtons[1] & 0x80) )
				s_bBtn2Flip = FALSE;
		}

		if( !s_bBtn3Flip )
		{
			if( js.rgbButtons[2] & 0x80 )
			{
				Toggle_Pause();
				s_bBtn3Flip = TRUE;
			}
		}
		else
		{
			if( !(js.rgbButtons[2] & 0x80) )
				s_bBtn3Flip = FALSE;
		}

		if( !s_bBtn4Flip )
		{
			if( js.rgbButtons[3] & 0x80 )
			{
				Toggle_SIOPatch();
				s_bBtn4Flip = TRUE;
			}
		}
		else
		{
			if( !(js.rgbButtons[3] & 0x80) )
				s_bBtn4Flip = FALSE;
		}

		if( bTestButton )
		{
			/* See if the repeat/skip options are being used for holding button down */
			if( g_nJoyRepeat )
			{
				if( js.rgbButtons[0] & 0x80)
				{
					/* Count down the number of repeat cycles and return button up for those */
					if( --s_nJoyCount > 0 )
						s_anTrig[sticknum] = 1;

					/* If we've counted the number of skip cycles, start over and send button down */
					if( s_nJoyCount < -g_nJoySkip )
					{
						s_nJoyCount = g_nJoyRepeat;
						s_anTrig[sticknum] = 0;
					}
				}
				else
				{
					/* Otherwise continue skipping, sending button up */
					s_nJoyCount = g_nJoyRepeat;
					s_anTrig[sticknum] = 1;
				}
			}
			else
			{
				/* No special handling, just return whether it is held down */
				if( js.rgbButtons[0] & 0x80)
					s_anTrig[sticknum] = 0;
				else
					s_anTrig[sticknum] = 1;
			}
		}
	}

	if( js.lX < 0 )
	{
		if( js.lY < 0 )
			stickval |= STICK_UL;
		if( js.lY > 0 )
			stickval |= STICK_LL;
		if( js.lY == 0 )
			stickval |= STICK_LEFT;
	}

	if( js.lX > 0 )
	{
		if( js.lY < 0 )
			stickval |= STICK_UR;
		if( js.lY > 0 )
			stickval |= STICK_LR;
		if( js.lY == 0 )
			stickval |= STICK_RIGHT;
	}

	if( js.lX == 0 )
	{
		if( js.lY < 0 )
			stickval |= STICK_FORWARD;

		if( js.lY > 0 )
			stickval |= STICK_BACK;

		if( js.lY == 0 )
			stickval |= STICK_CENTRE;
	}
	s_anStick[sticknum] = stickval;
	return 1;
}

/* In other Atari800 versions, keyboard handling is here. In Atari800Win I handle
   the actual key processing in the mainframe of the app, and by the time it gets
   here we already have the appropriate Atari key code (except for keycodes that 
   require multiple logic operations, like the s_anStick								*/
int Atari_Keyboard (void)
{
	if( g_nCurrentKey != AKEY_NONE )
	{
		if( GetAsyncKeyState( g_nCurrentVK ) >= 0 )
		{
			KEYPRESSED = FALSE;
			g_nCurrentKey = AKEY_NONE;
		}
	}

	if( g_nNewKey == AKEY_NONE )
		return g_nCurrentKey;

//	if( g_nNewVK & AKEY_PENDING )
//	{
//		g_nNewVK &= ~AKEY_PENDING;
//		return g_nCurrentKey;
//	}

	if (machine == Atari5200) 
	{
		switch ( g_nNewKey ) 
		{
		case AKEY_COLDSTART:
			g_nNewKey = AKEY_COLDSTART;
			break;
			
		case AKEY_WARMSTART:
			g_nNewKey = 0x29;		/* 5200 has no warmstart */	/* reset */
			break;
			
		case AKEY_STARTDWN:
			g_nNewKey = 0x39;		/* start */
			break;
			
		case AKEY_OPTIONDWN:		
			g_nNewKey = 0x31;		/* pause */
			break;
			
		case AKEY_0:
			g_nNewKey = 0x2b;
			break;
			
		case AKEY_1:
			g_nNewKey = 0x0f;
			break;
			
		case AKEY_2:
			g_nNewKey = 0x3f;
			break;
			
		case AKEY_3:
			g_nNewKey = 0x3d;
			break;
			
		case AKEY_4:
			g_nNewKey = 0x3b;
			break;
			
		case AKEY_5:
			g_nNewKey = 0x37;
			break;
			
		case AKEY_6:
			g_nNewKey = 0x35;
			break;
			
		case AKEY_7:
			g_nNewKey = 0x33;
			break;
			
		case AKEY_8:
			g_nNewKey = 0x2f;
			break;
			
		case AKEY_9:
			g_nNewKey = 0x2d;
			break;
			
		case AKEY_MINUS:
			g_nNewKey = 0x23;	/* # key on 5200 controller */
			break;
			
		case AKEY_ASTERISK:		/* * key on 5200 controller */
			g_nNewKey = 0x27;
			break;
			
		default:
			g_nNewKey = AKEY_NONE;
			break;
		}
	}

	if( g_nNewKey & SPECIAL_HANDLER_MASK )
	{
		int	keytemp = g_nNewKey;
		g_nNewKey = AKEY_NONE;

		if( g_nNewVK >= VK_NUMPAD0 && g_nNewVK <= VK_NUMPAD9 && !s_bKBJoystick )
			g_nNewKey = keytemp & ~SPECIAL_HANDLER_MASK;

		switch (keytemp) 
		{
			case	(AKEY_HELP | SPECIAL_HANDLER_MASK):
				keytemp &= ~SPECIAL_HANDLER_MASK;
				if( mach_xlxe )
					g_nNewKey = keytemp;
				break;
				
			case	(AKEY_F1 | SPECIAL_HANDLER_MASK):
			case	(AKEY_F2 | SPECIAL_HANDLER_MASK):
			case	(AKEY_F3 | SPECIAL_HANDLER_MASK):
			case	(AKEY_F4 | SPECIAL_HANDLER_MASK):
				keytemp &= ~SPECIAL_HANDLER_MASK;
//				if( SHIFT_KEY )
//					g_nNewKey |= AKEY_SHFT;
				if( machine == AtariXL )
					g_nNewKey = keytemp;
				break;
				
			case AKEY_COLDSTART:
				Coldstart();
				break;
				
			case AKEY_WARMSTART:
				Warmstart();
				break;
				
			case AKEY_EXIT:
				PostMessage( g_hMainWnd, WM_CLOSE, 0, 0L );
				break;
				
			case AKEY_BREAK:
				IRQST &= ~0x80;
				if (IRQEN & 0x80) {
					GenerateIRQ();
				}
				break;
				
			case AKEY_UI:
				monitor();	/* invoke monitor */
//#ifdef SUPPORT_UI
//				ui((UBYTE *)atari_screen);
//#endif
				break;
				
			case AKEY_PIL:
				if (pil_on)
					DisablePILL();
				else
					EnablePILL();
				break;
				
			default:
				break;
		}
		if( g_nNewKey == AKEY_NONE )
			g_nNewVK = 0;
	}

	if( g_nNewKey != AKEY_NONE && g_nCurrentKey == AKEY_NONE )
	{
		if( g_nNewKey != AKEY_NONE )
		{
			KEYPRESSED = TRUE;
			KBCODE = g_nCurrentKey = g_nNewKey;
			g_nCurrentVK = g_nNewVK;
			g_nNewKey = AKEY_NONE;
			g_nNewVK = 0;
			IRQST &= ~0x40;
			if (IRQEN & 0x40) 
			{
				GenerateIRQ();
			}
		}
	}
	return g_nCurrentKey;
}

void GetJoystickInput( void )
{
	int i;

	for( i = 0; i < MAX_ATARI_JOYPORTS; i++ )
	{
		if( s_alpJoyPorts[i] )
			ReadJoystickInput( i, TRUE );
		else
		{
			if( g_anSelectedInputDevice[i] == -1 )
			{
				s_anStick[i] = s_unKeystick;
				s_anTrig[i] = s_unKeytrig;
			}
			else
			{
				s_anStick[i] = 0xf;
				s_anTrig[i] = 1;
			}
		}
	}
}

int Atari_PORT( int num )
{
	if( !num )
		return (s_anStick[ 1 ] << 4) | s_anStick[ 0 ];
//		return( (s_anStick[ 1 ] & 0xF) << 4 | (s_anStick[ 0 ] & 0xF) );
	else
		return (s_anStick[ 3 ] << 4) | s_anStick[ 2 ];
}

int Atari_TRIG( int num )
{
	return s_anTrig[ num ];
}

int Atari_POT( int num )
{
	if( machine == Atari5200 )
	{
		if( num == 0 )
		{
			unsigned char pot0 = 120;
			if( !(s_anStick[0] & 0x04) )
				pot0 = 15;
			if( !(s_anStick[0] & 0x08) )
				pot0 = 220;
			return pot0;
		}
		
		if( num == 1)
		{
			unsigned char pot1 = 120;
			if( !(s_anStick[0] & 0x01) )
				pot1 = 15;
			if( !(s_anStick[0] & 0x02) )
				pot1 = 220;
			return pot1;
		}
	}
 	return 228;
}

int Atari_CONSOL (void)
{
  return s_nConsol;
}

static void Atari_Set_CONSOL( int newconsol )
{
	switch( newconsol )
	{
		case	AKEY_OPTIONDWN:	/* OPTION */
			s_nConsol &= 0x03;
			break;

		case	AKEY_SELECTDWN:	/* SELECT */
			s_nConsol &= 0x05;
			break;

		case	AKEY_STARTDWN:	/* START */
			s_nConsol &= 0x06;
			break;

		case	AKEY_OPTIONUP:
			s_nConsol |= 0x04;
			break;

		case	AKEY_SELECTUP:
			s_nConsol |= 0x02;
			break;

		case	AKEY_STARTUP:
			s_nConsol |= 0x01;
			break;
	}
}

BOOL ToggleKeyDown( WPARAM wp )
{
	if( wp >= VK_NUMPAD0 && wp <= VK_NUMPAD9 )
	{
		if( !s_bKBJoystick )
			return FALSE;

		if( wp == VK_NUMPAD0 )
		{
			s_unKeytrig = 0;
			return TRUE;
		}

		s_unKeystick &= s_anStickTable[ wp - VK_NUMPAD1 ];
		return TRUE;
	}

	switch( wp )
	{
		case	VK_F8:
			Toggle_FullSpeed();
			return TRUE;
			break;
			
		case	VK_F9:
			Toggle_Pause();
			return TRUE;
			break;
			
		case	VK_F11:
			Toggle_SIOPatch();
			return TRUE;
			break;
				
		case	VK_SHIFT:
			SHIFT_KEY = TRUE;
			return TRUE;
			break;

		case	VK_CONTROL:
			CTRL_KEY = TRUE;
			return TRUE;
			break;

		case	VK_F2:
			if( !SHIFT_KEY )	/* Console keys only work in non-shift mode */
				Atari_Set_CONSOL( AKEY_OPTIONDWN );
			break;

		case	VK_F3:
			if( !SHIFT_KEY )	/* Console keys only work in non-shift mode */
				Atari_Set_CONSOL( AKEY_SELECTDWN );
			break;

		case	VK_F4:
			if( !SHIFT_KEY )	/* Console keys only work in non-shift mode */
				Atari_Set_CONSOL( AKEY_STARTDWN );
			break;			
	}
	return FALSE;
}

BOOL ToggleKeyUp( WPARAM wp )
{
	if( wp >= VK_NUMPAD0 && wp <= VK_NUMPAD9 )
	{
		if( !s_bKBJoystick )
			return FALSE;

		if( wp == VK_NUMPAD0 )
		{
			s_unKeytrig = 1;
			return TRUE;
		}

		if( g_ulMiscStates & ATARI_STICK_RELEASE )
			s_unKeystick |= (~s_anStickTable[ wp - VK_NUMPAD1 ]) & 0xf; /* Turn bits for this key off */
		return TRUE;
	}

	switch( wp )
	{
		case	VK_SHIFT:
			SHIFT_KEY = FALSE;
			return TRUE;
			break;

		case	VK_CONTROL:
			CTRL_KEY = FALSE;
			return TRUE;
			break;

		case	VK_F2:
			if( !SHIFT_KEY )	/* Console keys only work in non-shift mode */
				Atari_Set_CONSOL( AKEY_OPTIONUP );
			break;

		case	VK_F3:
			if( !SHIFT_KEY )	/* Console keys only work in non-shift mode */
				Atari_Set_CONSOL( AKEY_SELECTUP );
			break;

		case	VK_F4:
			if( !SHIFT_KEY )	/* Console keys only work in non-shift mode */
				Atari_Set_CONSOL( AKEY_STARTUP );
			break;

		case	VK_F7:
			LaunchMonitor();
			break;
	}
	return FALSE;
}

static BOOL GetDIErrorString( HRESULT hResult, LPSTR lpszErrorBuff, DWORD cchError)
{
    DWORD  cLen;
    LPTSTR lpszError;
    TCHAR  szMsg[256];

    /* Check parameters */
    if (!lpszErrorBuff || !cchError)
    {
        /* Error, invalid parameters */
        return FALSE;
    }

    switch ( hResult )
    {
		case DI_OK:
			/* The operation completed successfully. This value is equal to the S_OK standard COM return value. */
			lpszError = TEXT("DI_OK");
			break;

#ifdef WIN_TRANSLATE_ERRORS

		case DI_BUFFEROVERFLOW:
			//The device buffer overflowed and some input was lost. This value is equal to the S_FALSE standard COM return value. 
			lpszError = TEXT("DI_BUFFEROVERFLOW");
			break;

		case DI_DOWNLOADSKIPPED:
			//The parameters of the effect were successfully updated, but the effect could not be downloaded because the associated device was not acquired in exclusive mode. 
			lpszError = TEXT("DI_DOWNLOADSKIPPED");
			break;

		case DI_EFFECTRESTARTED:
			//The effect was stopped, the parameters were updated, and the effect was restarted. 
			lpszError = TEXT("DI_EFFECTRESTARTED");
			break;

//		case DI_NOEFFECT:
			//The operation had no effect. This value is equal to the S_FALSE standard COM return value. 
//			lpszError = TEXT("DI_NOEFFECT");
//			break;

//		case DI_NOTATTACHED:
			//The device exists but is not currently attached. This value is equal to the S_FALSE standard COM return value. 
//			lpszError = TEXT("DI_NOTATTACHED");
//			break;
	
		case DI_POLLEDDEVICE:
			//The device is a polled device. As a result, device buffering will not collect any data and event notifications will not be signaled until the IDirectInputDevice2::Poll method is called. 
			lpszError = TEXT("DI_POLLEDDEVICE");
			break;

//		case DI_PROPNOEFFECT:
			//The change in device properties had no effect. This value is equal to the S_FALSE standard COM return value. 
//			lpszError = TEXT("DI_PROPNOEFFECT");
//			break;

		case DI_TRUNCATED:
			//The parameters of the effect were successfully updated, but some of them were beyond the capabilities of the device and were truncated to the nearest supported value. 
			lpszError = TEXT("DI_TRUNCATED");
			break;

		case DI_TRUNCATEDANDRESTARTED:
			//Equal to DI_EFFECTRESTARTED | DI_TRUNCATED.
			lpszError = TEXT("DI_TRUNCATEDANDRESTARTED");
			break;

		case DIERR_ACQUIRED:
			//The operation cannot be performed while the device is acquired. 
			lpszError = TEXT("DIERR_ACQUIRED");
			break;

		case DIERR_ALREADYINITIALIZED:
			//This object is already initialized 
			lpszError = TEXT("DIERR_ALREADYINITIALIZED");
			break;

		case DIERR_BADDRIVERVER:
			//The object could not be created due to an incompatible driver version or mismatched or incomplete driver components. 
			lpszError = TEXT("DIERR_BADDRIVERVER");
			break;

		case DIERR_BETADIRECTINPUTVERSION:
			//The application was written for an unsupported prerelease version of DirectInput. 
			lpszError = TEXT("DIERR_BETADIRECTINPUTVERSION");
			break;

		case DIERR_DEVICEFULL:
			//The device is full. 
			lpszError = TEXT("DIERR_DEVICEFULL");
			break;

		case DIERR_DEVICENOTREG:
			//The device or device instance is not registered with DirectInput. This value is equal to the REGDB_E_CLASSNOTREG standard COM return value. 
			lpszError = TEXT("DIERR_DEVICENOTREG");
			break;

		case DIERR_EFFECTPLAYING:
			//The parameters were updated in memory but were not downloaded to the device because the device does not support updating an effect while it is still playing. 
			lpszError = TEXT("DIERR_EFFECTPLAYING");
			break;

		case DIERR_HASEFFECTS:
			//The device cannot be reinitialized because there are still effects attached to it.
			lpszError = TEXT("DIERR_HASEFFECTS");
			break;

		case DIERR_GENERIC:
			//An undetermined error occurred inside the DirectInput subsystem. This value is equal to the E_FAIL standard COM return value. 
			lpszError = TEXT("DIERR_GENERIC");
			break;

		case DIERR_HANDLEEXISTS:
			//The device already has an event notification associated with it. This value is equal to the E_ACCESSDENIED standard COM return value. 
			lpszError = TEXT("DIERR_HANDLEEXISTS");
			break;

		case DIERR_INCOMPLETEEFFECT:
			//The effect could not be downloaded because essential information is missing. For example, no axes have been associated with the effect, or no type-specific information has been supplied.
			lpszError = TEXT("DIERR_INCOMPLETEEFFECT");
			break;

		case DIERR_INPUTLOST:
			//Access to the input device has been lost. It must be reacquired. 
			lpszError = TEXT("DIERR_INPUTLOST");
			break;

		case DIERR_INVALIDPARAM:
			//An invalid parameter was passed to the returning function, or the object was not in a state that permitted the function to be called. This value is equal to the E_INVALIDARG standard COM return value. 
			lpszError = TEXT("DIERR_INVALIDPARAM");
			break;

		case DIERR_MOREDATA:
			//Not all the requested information fitted into the buffer.
			lpszError = TEXT("DIERR_MOREDATA");
			break;

		case DIERR_NOAGGREGATION:
			//This object does not support aggregation. 
			lpszError = TEXT("DIERR_NOAGGREGATION");
			break;

		case DIERR_NOINTERFACE:
			//The specified interface is not supported by the object. This value is equal to the E_NOINTERFACE standard COM return value. 
			lpszError = TEXT("DIERR_NOINTERFACE");
			break;

		case DIERR_NOTACQUIRED:
			//The operation cannot be performed unless the device is acquired. 
			lpszError = TEXT("DIERR_NOTACQUIRED");
			break;

		case DIERR_NOTBUFFERED:
			//The device is not buffered. Set the DIPROP_BUFFERSIZE property to enable buffering.
			lpszError = TEXT("DIERR_NOTBUFFERED");
			break;

		case DIERR_NOTDOWNLOADED:
			//The effect is not downloaded.
			lpszError = TEXT("DIERR_NOTDOWNLOADED");
			break;

		case DIERR_NOTEXCLUSIVEACQUIRED:
			//The operation cannot be performed unless the device is acquired in DISCL_EXCLUSIVE mode. 
			lpszError = TEXT("DIERR_NOTEXCLUSIVEACQUIRED");
			break;

		case DIERR_NOTFOUND:
			//The requested object does not exist. 
			lpszError = TEXT("DIERR_NOTFOUND");
			break;

		case DIERR_NOTINITIALIZED:
			//This object has not been initialized. 
			lpszError = TEXT("DIERR_NOTINITIALIZED");
			break;

//		case DIERR_OBJECTNOTFOUND:
			//The requested object does not exist. 
//			lpszError = TEXT("DIERR_OBJECTNOTFOUND");
//			break;

		case DIERR_OLDDIRECTINPUTVERSION:
			//The application requires a newer version of DirectInput. 
			lpszError = TEXT("DIERR_OLDDIRECTINPUTVERSION");
			break;

//		case DIERR_OTHERAPPHASPRIO:
			//Another application has a higher priority level, preventing this call from succeeding. This value is equal to the E_ACCESSDENIED standard COM return value. This error can be returned when an application has only foreground access to a device but is attempting to acquire the device while in the background.
//			lpszError = TEXT("DIERR_OTHERAPPHASPRIO");
//			break;

		case DIERR_OUTOFMEMORY:
			//The DirectInput subsystem couldn't allocate sufficient memory to complete the call. This value is equal to the E_OUTOFMEMORY standard COM return value. 
			lpszError = TEXT("DIERR_OUTOFMEMORY");
			break;

//		case DIERR_READONLY:
//			//The specified property cannot be changed. This value is equal to the E_ACCESSDENIED standard COM return value. 
//			lpszError = TEXT("DIERR_READONLY");
//			break;

		case DIERR_UNSUPPORTED:
			//The function called is not supported at this time. This value is equal to the E_NOTIMPL standard COM return value. 
			lpszError = TEXT("DIERR_UNSUPPORTED");
			break;

		case E_PENDING:
			//Data is not yet available.
			lpszError = TEXT("E_PENDING");
			break;

#endif /*WIN_TRANSLATE_ERRORS*/

		default:
			// Unknown DI Error
			wsprintf (szMsg, "Error #%ld", (DWORD)(hResult & 0x0000FFFFL) );
			//wsprintf (szMsg, "Unknown Error #%ld", (DWORD)(hResult & 0x0000FFFFL));
			lpszError = szMsg;
			break;
	}
    // Copy DS Error string to buff
    cLen = strlen (lpszError);
    if (cLen >= cchError)
    {
        cLen = cchError - 1;
    }

    if (cLen)
    {
        strncpy (lpszErrorBuff, lpszError, cLen);
        lpszErrorBuff[cLen] = 0;
    }

    return TRUE;
}