/****************************************************************************
File    : CartridgeDlg.cpp
/*
@(#) #SY# Atari800Win
@(#) #IS# CCartridgeDlg implementation file
@(#) #BY# Richard Lawrence, Harry (modifications and code maintenance)
@(#) #LM# 20.02.2000
*/

/*
Copyright (c) 1998 Richard Lawrence

This program is free software; you can redistribute it and/or modify it under the terms 
of the GNU General Public License as published by the Free Software Foundation; either 
version 2 of the License, or (at your option) any later version. This program is 
distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. You should have received a copy of the GNU
General Public License along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "StdAfx.h"
#include "Atari800Win.h"
#include "WarningDlg.h"
#include "CartridgeDlg.h"

//#include "core.h"			// AtariWin core

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


/////////////////////////////////////////////////////////////////////////////
// Static objects

static char BASED_CODE s_szFilter[ LOADSTRING_STRING_SIZE ];


/////////////////////////////////////////////////////////////////////////////
// CCartridgeDlg dialog

BEGIN_MESSAGE_MAP(CCartridgeDlg, CDialog)
	//{{AFX_MSG_MAP(CCartridgeDlg)
	ON_BN_CLICKED(IDC_CARTRIDGE_BASIC_ROM, OnBasicRom)
	ON_BN_CLICKED(IDC_CARTRIDGE_OTHER_ROM, OnOtherRom)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/*========================================================
Method   : CCartridgeDlg::CCartridgeDlg
=========================================================*/
/* #FN#
   Standard constructor */
CCartridgeDlg::
CCartridgeDlg( CWnd* pParent /*=NULL*/ /* #IN# Pointer to parent window */ )
	: CDialog( CCartridgeDlg::IDD, pParent )
{
	//{{AFX_DATA_INIT(CCartridgeDlg)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT
	m_nCartState = 2;     /* Default is None */
	m_bReboot    = FALSE;
	LoadString( NULL, IDS_ROM_FILTER, s_szFilter, LOADSTRING_STRING_SIZE );
} /* #OF# CCartridgeDlg::CCartridgeDlg */

/*========================================================
Method   : CCartridgeDlg::DoDataExchange
=========================================================*/
/* #FN#
   Dynamic Data Exchange (not using) */
void
/* #AS#
   Nothing */
CCartridgeDlg::
DoDataExchange( CDataExchange *pDX /* #IN# Pointer to CDataExchange object */ )
{
	CDialog::DoDataExchange( pDX );
	//{{AFX_DATA_MAP(CCartridgeDlg)
		// NOTE: the ClassWizard will add DDX and DDV calls here
	//}}AFX_DATA_MAP
} /* #OF# CCartridgeDlg::DoDataExchange */


/////////////////////////////////////////////////////////////////////////////
// CCartridgeDlg implementation

/*========================================================
Method   : CCartridgeDlg::SetDlgState
=========================================================*/
/* #FN#
   Sets up the state of the dialog controls */
void
/* #AS#
   Nothing */
CCartridgeDlg::
SetDlgState()
{
	CButton	*pButton = NULL;
	BOOL     bCurrentNone;

	/* Set BASIC ROM path edit window */
	SetDlgItemText( IDC_CARTRIDGE_BASIC_EDIT, atari_basic_filename );
	/* Set Other ROM path edit window */
	ReadRegBinary( NULL, REG_OTHER_ROM, m_szOtherROM, MAX_PATH, FALSE );
	SetDlgItemText( IDC_CARTRIDGE_OTHER_EDIT, m_szOtherROM );

	/* The BASIC cartridge is not used */
	if( !(default_system == 1/*800 OS-A*/ || default_system == 2/*800 OS-B*/) )
	{
		pButton = (CButton*)GetDlgItem( IDC_CARTRIDGE_BASIC );
		ASSERT(pButton);
		pButton->EnableWindow( FALSE );
		/* We need case insensitive comparison here */
		bCurrentNone =
			/* BASIC is the current ROM or there is no other ROM */
			!_stricmp( current_rom, atari_basic_filename ) ||
			!strcmp( m_szOtherROM, "None" );
	}
	else
		bCurrentNone =
			/* There is no other ROM and
			   BASIC isn't the current ROM or OPTION button is holded */
			!strcmp( m_szOtherROM, "None" ) &&
			(_stricmp( current_rom, atari_basic_filename ) || hold_option);

	if( bCurrentNone )
	{
		/* That isn't any good idea to verify these values here. We need
		   to change CHardwareDlg methods where the machine type is
		   selected rather. The only positive thing of that strange code
		   is the auto-verification of corrupted registry values */
		strcpy( current_rom, "None" );
		cart_type = NO_CART;
		if( rom_inserted )
			m_bReboot = TRUE;
		/* There may be wrong values in the registry, fix it now */
		WriteRegString( NULL, REG_CURRENT_ROM, current_rom );
		WriteRegDWORD ( NULL, REG_CART_TYPE, cart_type );
	}

	if( !strcmp( current_rom, "None" ) )
		rom_inserted = FALSE;

	/* Set appropriate Cart Type radio button */
	if( rom_inserted )
		m_nCartState = _stricmp( current_rom, atari_basic_filename ) ? 1 : 0;
	CheckRadioButton( IDC_CARTRIDGE_BASIC, IDC_CARTRIDGE_NONE, IDC_CARTRIDGE_BASIC + m_nCartState );

	if( hold_option )
	{
		pButton = (CButton*)GetDlgItem( IDC_CARTRIDGE_HOLD_OPTION );
		ASSERT(pButton);
		pButton->SetCheck( 1 );
	}
} /* #OF# CCartridgeDlg::SetDlgState */


/////////////////////////////////////////////////////////////////////////////
// CCartridgeDlg message handlers

/*========================================================
Method   : CCartridgeDlg::OnInitDialog
=========================================================*/
/* #FN#
   Performs special processing when the dialog box is initialized */
BOOL
/* #AS#
   TRUE unless you set the focus to a control */
CCartridgeDlg::
OnInitDialog() 
{
	CDialog::OnInitDialog();

	SetDlgState();

	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
} /* #OF# CCartridgeDlg::OnInitDialog */

/*========================================================
Method   : CCartridgeDlg::OnBasicRom
=========================================================*/
/* #FN#
   Allows to choose new BASIC ROM image using standard file dialog */
void
/* #AS#
   Nothing */
CCartridgeDlg::
OnBasicRom() 
{
	CString	strBasicROM;
	CFileDialog	fdlgROMImage( TRUE, NULL, NULL, OFN_EXPLORER | OFN_FILEMUSTEXIST, s_szFilter, this );

	fdlgROMImage.m_ofn.lpstrTitle = "Select ROM for BASIC cart";
	if( fdlgROMImage.DoModal() == IDOK )
	{
		unsigned long ulCRC;
		int nCFResult;

		strBasicROM = fdlgROMImage.GetPathName();
		SetDlgItemText( IDC_CARTRIDGE_BASIC_EDIT, strBasicROM.GetBuffer( 0 ) );
		ulCRC = CheckFile( strBasicROM.GetBuffer( 0 ), &nCFResult );
		if( !nCFResult && (ulCRC != 2190982779L ) && !(g_ulDontShowFlags & DONT_SHOW_BASIC_WARN) )
		{
			CWarningDlg	dlgWarning;
			char szMsg[ LOADSTRING_STRING_SIZE ];
			LoadString( NULL, IDS_WARNING_ROM_CORRUPT, szMsg, LOADSTRING_STRING_SIZE );
			dlgWarning.csWarnText = szMsg;
			dlgWarning.iWarnBit   = DONT_SHOW_BASIC_WARN;
			dlgWarning.DoModal();
		}
		strBasicROM.ReleaseBuffer();
	}
} /* #OF# CCartridgeDlg::OnBasicRom */

/*========================================================
Method   : CCartridgeDlg::OnOtherRom
=========================================================*/
/* #FN#
   Allows to choose new Other ROM image using standard file dialog */
void
/* #AS#
   Nothing */
CCartridgeDlg::
OnOtherRom()
{
	CString strOtherROM;
	char    szCurDir[ MAX_PATH ];
	char    szCurRom[ MAX_PATH ];
	CFileDialog	fdlgROMImage( TRUE, NULL, NULL, OFN_EXPLORER | OFN_FILEMUSTEXIST, s_szFilter, this );

	fdlgROMImage.m_ofn.lpstrTitle = "Select ROM to use as inserted cart";
	GetDlgItemText( IDC_CARTRIDGE_OTHER_EDIT, szCurRom, MAX_PATH );
	GetCurrentDirectory( MAX_PATH, szCurDir );

	if( strcmp( szCurRom, "None" ) && strcmp( szCurRom, "Off" ) )
	{
		for( int i = strlen( szCurRom ); i > 0 && szCurRom[ i ] != '\\'; i-- );
		if( i > 0 )
			szCurRom[ i ] = '\0';
		/* Make the last used directory as default */
		SetCurrentDirectory( szCurRom );
	}
	if( fdlgROMImage.DoModal() == IDOK )
	{
		strOtherROM = fdlgROMImage.GetPathName();
		SetDlgItemText( IDC_CARTRIDGE_OTHER_EDIT, strOtherROM.GetBuffer( 0 ) );
		strOtherROM.ReleaseBuffer();
	}
	SetCurrentDirectory( szCurDir );
} /* #OF# CCartridgeDlg::OnOtherRom */

/*========================================================
Method   : CCartridgeDlg::OnOK
=========================================================*/
/* #FN#
   Handles the pressing OK button event */
void
/* #AS#
   Nothing */
CCartridgeDlg::
OnOK()
{
	char     szNewROM[ MAX_PATH ];
	int      nNewCartState;
	int      nNewHoldOption;
	CButton *pButton = NULL;

	CDialog::OnOK();

	/* Get cartridge state the user was set */
	nNewCartState = GetCheckedRadioButton( IDC_CARTRIDGE_BASIC, IDC_CARTRIDGE_NONE ) - IDC_CARTRIDGE_BASIC;

	GetDlgItemText( IDC_CARTRIDGE_BASIC_EDIT, szNewROM, MAX_PATH );
	if( *szNewROM == '\0' )
		strcpy( szNewROM, "None" );
	if( _stricmp( szNewROM, atari_basic_filename ) )
	{
		strcpy( atari_basic_filename, szNewROM );
		if( nNewCartState == 0 )
			m_bReboot = TRUE;
		WriteRegString( NULL, REG_BASIC_ROM, atari_basic_filename );
	}

	GetDlgItemText( IDC_CARTRIDGE_OTHER_EDIT, szNewROM, MAX_PATH );
	if( *szNewROM == '\0' )
		strcpy( szNewROM, "None" );
	if( _stricmp( szNewROM, m_szOtherROM ) )
	{
		if( nNewCartState == 1 )
		{
			strcpy( current_rom, szNewROM );
			m_bReboot = TRUE;
		}
		strcpy( m_szOtherROM, szNewROM );
		WriteRegString( NULL, REG_OTHER_ROM, m_szOtherROM );
	}

	if( nNewCartState != m_nCartState || m_bReboot )
	{
		m_bReboot = TRUE;
		if( nNewCartState == 0 )
		{
			if( rom_inserted )
				Remove_ROM();
			/* strcpy will be called within Insert_ROM function */
			/* strcpy( current_rom, atari_basic_filename ); */
			if( !Insert_8K_ROM( atari_basic_filename ) )
			{
				LoadString( NULL, IDS_ERROR_LOADING_ROM, g_tszErrorString, LOADSTRING_STRING_SIZE );
				if( strcmp( atari_basic_filename, "None" ) )
					strcat( g_tszErrorString, atari_basic_filename );
				MessageBox( g_tszErrorString, "Atari800Win", MB_ICONSTOP | MB_OK );

				strcpy( current_rom, "None" );
				cart_type = NO_CART;
			}
		}
		if( nNewCartState == 1 )
		{
			CFile cfFile;
			int	  nLength =  0;
			int   nResult = -1;

			if( rom_inserted )
				Remove_ROM();
			/* strcpy will be called within Insert_ROM functions */
			/* strcpy( current_rom, m_szOtherROM ); */
			if( cfFile.Open( m_szOtherROM, CFile::modeRead | CFile::typeBinary, NULL ) )
			{
				nLength = cfFile.GetLength();
				cfFile.Close();
			}
			cart_type = NO_CART;
			if( nLength < 8193 )
			{
				nResult = Insert_8K_ROM( m_szOtherROM );
				if( nResult )
					cart_type = NORMAL8_CART;
			}
			else if( nLength < 16385 )
			{
				nResult = Insert_16K_ROM( m_szOtherROM );
				if( nResult )
					cart_type = NORMAL16_CART;
			}
			else if( nLength < 32769 )
			{
				nResult = Insert_32K_5200ROM( m_szOtherROM );
				if( nResult )
					cart_type = AGS32_CART;
			}
			if( nResult <= 0 )
			{
				LoadString( NULL,
					(nResult == -1) ? IDS_UNKNOWN_ROM_FORMAT : IDS_ERROR_LOADING_ROM,
					g_tszErrorString, LOADSTRING_STRING_SIZE );
				MessageBox( g_tszErrorString, "Atari800Win", MB_ICONEXCLAMATION | MB_OK );
				strcpy( current_rom, "None" );
			}
		}
		if( nNewCartState == 2 )
		{
			strcpy( current_rom, "None" );
			cart_type = NO_CART;
		}
		WriteRegString( NULL, REG_CURRENT_ROM, current_rom );
		WriteRegDWORD ( NULL, REG_CART_TYPE, cart_type );
	}

	pButton = (CButton*)GetDlgItem( IDC_CARTRIDGE_HOLD_OPTION );
	ASSERT(pButton);
	nNewHoldOption = pButton->GetCheck() ? 1 : 0;
	if( nNewHoldOption != hold_option )
	{
		hold_option = nNewHoldOption;
		m_bReboot = TRUE;
		WriteRegDWORD( NULL, REG_HOLD_OPTION, hold_option );
	}

	if( m_bReboot )
	{
		Atari_ReInit();
	}
} /* #OF# CCartridgeDlg::OnOK */
