/*
Copyright (c) 1998 Richard Lawrence

This program is free software; you can redistribute it and/or modify it under the terms 
of the GNU General Public License as published by the Free Software Foundation; either 
version 2 of the License, or (at your option) any later version. This program is 
distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. You should have received a copy of the GNU
General Public License along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

// CGraphicsSmallDlg.cpp : implementation file
//

#include "StdAfx.h"
#include "Atari800Win.h"
#include "Graphics.h"
#include "GraphicsSmallDlg.h"
#include "WarningDlg.h"

//#include "core.h"			// AtariWin core

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


/////////////////////////////////////////////////////////////////////////////
// Static objects

static ULONG m_ulScreenMode;


/////////////////////////////////////////////////////////////////////////////
// CGraphicsSmallDlg dialog


CGraphicsSmallDlg::CGraphicsSmallDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CGraphicsSmallDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CGraphicsSmallDlg)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT
}


void CGraphicsSmallDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CGraphicsSmallDlg)
		// NOTE: the ClassWizard will add DDX and DDV calls here
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CGraphicsSmallDlg, CDialog)
	//{{AFX_MSG_MAP(CGraphicsSmallDlg)
	ON_BN_CLICKED(IDC_NODDRAW_SMALL, OnNoddrawSmall)
	ON_CBN_SELCHANGE(IDC_GRAPHICS_COMBO_SMALL, OnSelchangeGraphicsComboSmall)
	ON_BN_CLICKED(IDC_FORCE_VSYNC_SMALL, OnForceVsyncSmall)
	ON_BN_CLICKED(IDC_SMALL_LOCALMEM, OnSmallLocalmem)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CGraphicsSmallDlg message handlers

void CGraphicsSmallDlg::SetDlgState( void )
{
	int iSelect = 0;
	HANDLE	icon = NULL;
	CStatic* iconDisp = (CStatic *)GetDlgItem( IDC_ICON_GRAPHICS_SMALL );

	if( m_ulScreenMode & SM_MODE_FULL )
	{
		icon = AfxGetApp()->LoadIcon( IDI_FULL_OVERSCAN );
		iSelect = 5;		//Default case (800 x 600)

		if( m_ulScreenMode & SM_FRES_320_200 )
		{
			icon = AfxGetApp()->LoadIcon( IDI_NO_OVERSCAN );
			iSelect = 0;
		}
		if( m_ulScreenMode & SM_FRES_320_240 )
		{
			icon = AfxGetApp()->LoadIcon( IDI_VERT_OVERSCAN );
			iSelect = 1;
		}
		if( m_ulScreenMode & SM_FRES_512_384 )
		{
			iSelect = 2;
		}
		if( m_ulScreenMode & SM_FRES_640_400 )
		{
			icon = AfxGetApp()->LoadIcon( IDI_NO_OVERSCAN );
			iSelect = 3;
		}
		if( m_ulScreenMode & SM_FRES_640_480 )
		{
			icon = AfxGetApp()->LoadIcon( IDI_VERT_OVERSCAN );
			iSelect = 4;
		}
		if( m_ulScreenMode & SM_FRES_1024_768 )
		{
			iSelect = 6;
		}
	}
	else
	{
		if( m_ulScreenMode & SM_WRES_NORMAL )
		{
			icon = AfxGetApp()->LoadIcon( IDI_FULL_OVERSCAN );
			iSelect = 7;
		}
		else
		{
			icon = AfxGetApp()->LoadIcon( IDI_FULL_OVERSCAN );
			iSelect = 8;
		}
	}

	cbCheckBox = (CButton *)GetDlgItem( IDC_NODDRAW_SMALL );
	ASSERT( cbCheckBox );
	if( m_ulScreenMode & SM_OPTN_USE_GDI )
		cbCheckBox->SetCheck( 1 );
	else
		cbCheckBox->SetCheck( 0 );

	cbCheckBox = (CButton *)GetDlgItem( IDC_FORCE_VSYNC_SMALL );
	ASSERT( cbCheckBox );
	cbCheckBox->SetCheck(
		 m_ulScreenMode & SM_OPTN_DDBLT_WAIT ? 1 :
		(m_ulScreenMode & SM_OPTN_DDVBL_WAIT ? 2 : 0) );

	ComboBox->SetCurSel( iSelect );
	if( icon && iconDisp )
		iconDisp->SetIcon( (HICON)icon );

	CButton *cbButton = (CButton *)GetDlgItem( IDC_SMALL_LOCALMEM );
	ASSERT( cbButton );	
	if( m_ulScreenMode & SM_OPTN_USE_VIDEOMEM )
		cbButton->SetCheck( 1 );
	else
		cbButton->SetCheck( 0 );
}

BOOL CGraphicsSmallDlg::OnInitDialog() 
{
	CDialog::OnInitDialog();

	m_ulScreenMode = g_ulScreenMode;

	ComboBox = (CComboBox *)GetDlgItem( IDC_GRAPHICS_COMBO_SMALL );
	
	if( !ComboBox )
		return FALSE;

	SetDlgState();

	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void CGraphicsSmallDlg::OnSelchangeGraphicsComboSmall() 
{
	switch( ComboBox->GetCurSel() )
	{
		case	0:
			{
				int result = IDOK;
				if( !(g_ulDontShowFlags & DONT_SHOW_MENU_WARN) )
				{
					CWarningDlg Warning;
					char	message[LOADSTRING_STRING_SIZE];
					LoadString( NULL, IDS_WARNING_NO_MENU, message, LOADSTRING_STRING_SIZE );
					Warning.csWarnText = message;
					Warning.iWarnBit = DONT_SHOW_MENU_WARN;
					result = Warning.DoModal();
				}
				if( result == IDOK )
				{
					m_ulScreenMode &= ~(SM_FRES_MASK | SM_MODE_MASK);
					m_ulScreenMode |= SM_FRES_320_200 | SM_MODE_FULL;
				}
				break;
			}

		case	1:
			m_ulScreenMode &= ~(SM_FRES_MASK | SM_MODE_MASK);
			m_ulScreenMode |= SM_FRES_320_240 | SM_MODE_FULL;
			break;

		case	2:
			m_ulScreenMode &= ~(SM_FRES_MASK | SM_MODE_MASK);
			m_ulScreenMode |= SM_FRES_512_384 | SM_MODE_FULL;
			break;

		case	3:
			{
				int result = IDOK;
				if( !(g_ulDontShowFlags & DONT_SHOW_MENU_WARN) )
				{
					CWarningDlg Warning;
					char	message[LOADSTRING_STRING_SIZE];
					LoadString( NULL, IDS_WARNING_NO_MENU, message, LOADSTRING_STRING_SIZE );
					Warning.csWarnText = message;
					Warning.iWarnBit = DONT_SHOW_MENU_WARN;
					Warning.DoModal();
				}
				if( result == IDOK )
				{
					m_ulScreenMode &= ~(SM_FRES_MASK | SM_MODE_MASK);
					m_ulScreenMode |= SM_FRES_640_400 | SM_MODE_FULL;
				}
				break;
			}

		case	4:
			m_ulScreenMode &= ~(SM_FRES_MASK | SM_MODE_MASK);
			m_ulScreenMode |= SM_FRES_640_480 | SM_MODE_FULL;
			break;

		case	5:
			m_ulScreenMode &= ~(SM_FRES_MASK | SM_MODE_MASK);
			m_ulScreenMode |= SM_FRES_800_600 | SM_MODE_FULL;
			break;

		case	6:
			m_ulScreenMode &= ~(SM_FRES_MASK | SM_MODE_MASK);
			m_ulScreenMode |= SM_FRES_1024_768 | SM_MODE_FULL;
			break;

		case	7:
			m_ulScreenMode &= ~(SM_WRES_MASK | SM_MODE_MASK);
			m_ulScreenMode |= SM_WRES_NORMAL | SM_MODE_WIND;
			break;

		case	8:
			{
				int result = IDOK;
				if( !(g_ulDontShowFlags & DONT_SHOW_GRAPHICS_WARN ) )
				{
					CWarningDlg	Warning;
					char	message[LOADSTRING_STRING_SIZE];

					LoadString( NULL, IDS_WARNING_GRAPHICS_WINDOW, message, LOADSTRING_STRING_SIZE );
					Warning.csWarnText = message;
					Warning.iWarnBit = DONT_SHOW_GRAPHICS_WARN;
					result = Warning.DoModal();
				}

				if( result == IDOK )
				{
					m_ulScreenMode &= ~(SM_WRES_MASK | SM_MODE_MASK);
					m_ulScreenMode |= SM_WRES_STRETCH | SM_MODE_WIND;
				}
			}
			break;

		default:
			m_ulScreenMode &= ~(SM_FRES_MASK | SM_MODE_MASK);
			m_ulScreenMode |= SM_FRES_640_480 | SM_MODE_FULL;
			break;
	}
	SetDlgState();
}

void CGraphicsSmallDlg::OnNoddrawSmall() 
{
	cbCheckBox = (CButton *)GetDlgItem( IDC_NODDRAW_SMALL );
	if( cbCheckBox )
	{
		if( cbCheckBox->GetCheck() )
			m_ulScreenMode |= SM_OPTN_USE_GDI;
		else
			m_ulScreenMode &= ~SM_OPTN_USE_GDI;
	}
	SetDlgState();
}

void CGraphicsSmallDlg::OnSmallLocalmem() 
{
	CButton *cbButton = (CButton *)GetDlgItem( IDC_SMALL_LOCALMEM );
	ASSERT( cbButton );	

	if( cbButton->GetCheck() )
		m_ulScreenMode |= SM_OPTN_USE_VIDEOMEM;
	else
		m_ulScreenMode &= ~SM_OPTN_USE_VIDEOMEM;
}

void CGraphicsSmallDlg::OnForceVsyncSmall() 
{
	CButton *cbButton = (CButton *)GetDlgItem( IDC_FORCE_VSYNC_SMALL );
	ASSERT( cbButton );
	m_ulScreenMode &= ~(SM_OPTN_DDBLT_WAIT | SM_OPTN_DDVBL_WAIT);
	if( cbButton->GetCheck() )
		m_ulScreenMode |= (cbButton->GetCheck() == 1 ? SM_OPTN_DDBLT_WAIT : SM_OPTN_DDVBL_WAIT);
	SetDlgState();
}

void CGraphicsSmallDlg::OnOK() 
{
	CDialog::OnOK();

	if( m_ulScreenMode != g_ulScreenMode )
	{
		ULONG	savedmode = g_ulScreenMode;

		g_ulScreenMode = m_ulScreenMode;
		if( !InitialiseScreen() )
		{
			g_ulScreenMode = savedmode;
			InitialiseScreen();
		}
		WriteRegDWORD( NULL, REG_DDRAW_MODE, g_ulScreenMode );
	}
}
