/****************************************************************************
File    : HardwareDlg.cpp
/*
@(#) #SY# Atari800Win
@(#) #IS# CHardwareDlg implementation file
@(#) #BY# Richard Lawrence, Harry (modifications and code maintenance)
@(#) #LM# 22.02.2000
*/

/*
Copyright (c) 1998 Richard Lawrence

This program is free software; you can redistribute it and/or modify it under the terms 
of the GNU General Public License as published by the Free Software Foundation; either 
version 2 of the License, or (at your option) any later version. This program is 
distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. You should have received a copy of the GNU
General Public License along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "StdAfx.h"
#include "Atari800Win.h"
#include "WarningDlg.h"
#include "HardwareDlg.h"

//#include "core.h"			// AtariWin core

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


/////////////////////////////////////////////////////////////////////////////
// Static objects

static char BASED_CODE s_szFilter[] = "ROM Images (*.rom)|*.rom|All Files (*.*)|*.*||";

static unsigned int s_anWarnBits[ MACHINE_TYPES ] =
{
	DONT_SHOW_OSA_WARN,
	DONT_SHOW_OSB_WARN,
	DONT_SHOW_XL_WARN,
	DONT_SHOW_XL_WARN,
	DONT_SHOW_XL_WARN,
	DONT_SHOW_XL_WARN,
	DONT_SHOW_5200_WARN
};

static unsigned long s_alFileCRC[ MACHINE_TYPES ] =
{
	0L,
	3252116993L,
	3764596111L,
	3764596111L,
	3764596111L,
	3764596111L,
	3182898204L
};

static char *s_apszRomNames[ MACHINE_TYPES ] =
{
	atari_osa_filename,
	atari_osb_filename,
	atari_xlxe_filename,
	atari_xlxe_filename,
	atari_xlxe_filename,
	atari_xlxe_filename,
	atari_5200_filename
};

static char s_acNewRomNames[ MACHINE_TYPES ][ MAX_PATH ];


/////////////////////////////////////////////////////////////////////////////
// CHardwareDlg dialog

BEGIN_MESSAGE_MAP(CHardwareDlg, CDialog)
	//{{AFX_MSG_MAP(CHardwareDlg)
	ON_NOTIFY(UDN_DELTAPOS, IDC_REFRESH_SPIN, OnDeltaposRefreshSpin)
	ON_EN_KILLFOCUS(IDC_REFRESH_RATE, OnKillfocusRefreshRate)
	ON_EN_KILLFOCUS(IDC_ROM_EDIT, OnKillfocusRomEdit)
	ON_CBN_SELCHANGE(IDC_MACHINE_SELECTOR, OnSelchangeMachineSelector)
	ON_CBN_SELCHANGE(IDC_ARTIFACT_MODE, OnSelchangeArtifactMode)
	ON_BN_CLICKED(IDC_HARDWARE_NTSC, OnNtsc)
	ON_BN_CLICKED(IDC_HARDWARE_PAL, OnPal)
	ON_BN_CLICKED(IDC_ROMFILE, OnRomfile)
	ON_BN_CLICKED(IDC_ENABLE_ROM, OnEnableRom)
	ON_BN_CLICKED(IDC_NEW_BINLOAD, OnNewBinload)
	ON_BN_CLICKED(IDC_FULL_SPEED, OnFullSpeed)
	ON_BN_CLICKED(IDC_ENABLE_SIO, OnEnableSio)
	ON_BN_CLICKED(IDC_C000_RAM, OnC000Ram)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/*========================================================
Method   : CHardwareDlg::CHardwareDlg
=========================================================*/
/* #FN#
   Standard constructor */
CHardwareDlg::
CHardwareDlg( CWnd *pParent /*=NULL*/ /* #IN# Pointer to parent window */ )
	: CDialog( CHardwareDlg::IDD, pParent )
{
	//{{AFX_DATA_INIT(CHardwareDlg)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT
} /* #OF# CHardwareDlg::CHardwareDlg */

/*========================================================
Method   : CHardwareDlg::DoDataExchange
=========================================================*/
/* #FN#
   Dynamic Data Exchange (not using) */
void
/* #AS#
   Nothing */
CHardwareDlg::
DoDataExchange( CDataExchange *pDX /* #IN# Pointer to CDataExchange object */ )
{
	CDialog::DoDataExchange( pDX );
	//{{AFX_DATA_MAP(CHardwareDlg)
		// NOTE: the ClassWizard will add DDX and DDV calls here
	//}}AFX_DATA_MAP
} /* #OF# CHardwareDlg::DoDataExchange */


/////////////////////////////////////////////////////////////////////////////
// CHardwareDlg implementation

/*========================================================
Method   : CHardwareDlg::SetDlgState
=========================================================*/
/* #FN#
   Sets up the state of the dialog controls */
void
/* #AS#
   Nothing */
CHardwareDlg::
SetDlgState()
{
	char       szDigits[ 64 ];
	CComboBox *pCombo  = NULL;
	CButton	  *pButton = NULL;

	pCombo = (CComboBox*)GetDlgItem( IDC_ARTIFACT_MODE );
	ASSERT(pCombo);
	pCombo->SetCurSel( m_nGlobalArtifMode );

	if( m_nSystem >= 0 && m_nSystem <= MACHINE_TYPES  )
	{
		pCombo = (CComboBox*)GetDlgItem( IDC_MACHINE_SELECTOR );
		ASSERT(pCombo);
		pCombo->SetCurSel( m_nSystem - 1 );
		SetDlgItemText( IDC_ROM_EDIT, s_acNewRomNames[ m_nSystem - 1 ] );
	}
	else
		SetDlgItemText( IDC_ROM_EDIT, "Bad machine type, please reselect" );

	CheckRadioButton( IDC_HARDWARE_NTSC, IDC_HARDWARE_PAL,
		m_nDefaultTVMode == 1 ? IDC_HARDWARE_PAL : IDC_HARDWARE_NTSC );

	itoa( m_nRefreshRate, szDigits, 10 );
	SetDlgItemText( IDC_REFRESH_RATE, szDigits );

	/* Set up check buttons states */
	if( m_ulMiscStates & ATARI_FULL_SPEED )
	{
		pButton = (CButton*)GetDlgItem( IDC_FULL_SPEED );
		ASSERT(pButton);
		pButton->SetCheck( 1 );
	}

	pButton = (CButton*)GetDlgItem( IDC_ENABLE_ROM );
	ASSERT(pButton);
	if( m_nEnableRomPatches )
	{
		pButton->SetCheck( 1 );

		pButton = (CButton*)GetDlgItem( IDC_ENABLE_SIO );
		ASSERT(pButton);
		pButton->EnableWindow( TRUE );

		if( m_nEnableSIOPatch )
			pButton->SetCheck( 1 );
	}
	else /* ROM patch == FALSE automatically disables SIO_PATCH */
	{
		pButton->SetCheck( 0 );

		pButton = (CButton*)GetDlgItem( IDC_ENABLE_SIO );
		ASSERT(pButton);
		pButton->SetCheck( 0 );
		pButton->EnableWindow( FALSE );
		m_nEnableSIOPatch = 0; /* turn off SIO patch */
	}

	if( m_nEnableC000Ram )
	{
		pButton = (CButton*)GetDlgItem( IDC_C000_RAM );
		ASSERT(pButton);
		pButton->SetCheck( 1 );
	}

	if( m_nUseNewBinload )
	{
		pButton = (CButton*)GetDlgItem( IDC_NEW_BINLOAD );
		ASSERT(pButton);
		pButton->SetCheck( 1 );
	}
#ifndef USE_NEW_BINLOAD
	GetDlgItem( IDC_NEW_BINLOAD )->EnableWindow( FALSE );
#endif
} /* #OF# CHardwareDlg::SetDlgState */


/////////////////////////////////////////////////////////////////////////////
// CHardwareDlg message handlers

/*========================================================
Method   : CHardwareDlg::OnInitDialog
=========================================================*/
/* #FN#
   Performs special processing when the dialog box is initialized */
BOOL
/* #AS#
   TRUE unless you set the focus to a control */
CHardwareDlg::
OnInitDialog()
{
	CDialog::OnInitDialog();

	switch( default_system )
	{
		case 1:
			m_nSystem = 1;		// OS-A
			break;
		case 2:
			m_nSystem = 2;		// OS-B
			break;
		case 3:
			m_nSystem = 3;		// 800 XL
			break;
		case 4:
			m_nSystem = 4;		// 130 XE
			break;
		case 5:
			if( Ram256 == 2 )
				m_nSystem = 5;	// 320 XE Compy
			else
				m_nSystem = 6;	// 320 XE Rambo
			break;
		case 6:
			m_nSystem = 7;		// 5200
			break;
		default:
			m_nSystem = 2;		// OS-B
			break;
	}

	if( default_tv_mode < 1 || default_tv_mode > 2 )
		default_tv_mode = 1;

	m_nOrigSystem       = m_nSystem;
	m_nGlobalArtifMode  = global_artif_mode;
	m_nDefaultTVMode    = default_tv_mode;
	m_nRefreshRate      = refresh_rate;
	m_nEnableRomPatches = enable_rom_patches;
	m_nEnableSIOPatch   = enable_sio_patch;
	m_nEnableC000Ram    = enable_c000_ram;
	m_nUseNewBinload    = use_new_binload;
	m_ulMiscStates      = g_ulMiscStates;

	for( int i = 0; i < MACHINE_TYPES; i++ )
		strcpy( s_acNewRomNames[ i ], s_apszRomNames[ i ] );

	SetDlgState();

	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
} /* #OF# CHardwareDlg::OnInitDialog */

/*========================================================
Method   : CHardwareDlg::OnSelchangeArtifactMode
=========================================================*/
/* #FN#
   Sets the state of object regarding to Artifacting Mode combo */
void
/* #AS#
   Nothing */
CHardwareDlg::
OnSelchangeArtifactMode()
{
	CComboBox *pCombo = (CComboBox*)GetDlgItem( IDC_ARTIFACT_MODE );
	ASSERT(pCombo);

	m_nGlobalArtifMode = pCombo->GetCurSel();
} /* #OF# CHardwareDlg::OnSelchangeArtifactMode */

/*========================================================
Method   : CHardwareDlg::OnSelchangeMachineSelector
=========================================================*/
/* #FN#
	Sets the state of object regarding to Machine Type combo */
void
/* #AS#
   Nothing */
CHardwareDlg::
OnSelchangeMachineSelector()
{
	CComboBox *pCombo = (CComboBox*)GetDlgItem( IDC_MACHINE_SELECTOR );
	ASSERT(pCombo);

	m_nSystem = pCombo->GetCurSel() + 1;
	SetDlgState();
} /* #OF# CHardwareDlg::OnSelchangeMachineSelector */

/*========================================================
Method   : CHardwareDlg::OnSelchangeMachineSelector
=========================================================*/
/* #FN#
	Sets the state of object regarding to Machine Type combo */
int
/* #AS#
   IDOK or IDCANCEL */
CHardwareDlg::
OnKillfocusRomEdit()
{
	char  szBuffer[ MAX_PATH ];
	ULONG ulCRC;
	int   nCFResult;
	int   nResult = IDOK;

	ASSERT(m_nSystem >= 0 && m_nSystem <= MACHINE_TYPES);

	GetDlgItemText( IDC_ROM_EDIT, szBuffer, MAX_PATH );
	ulCRC = CheckFile( szBuffer, &nCFResult );

	if( ulCRC && !nCFResult && (ulCRC != s_alFileCRC[ m_nSystem - 1 ]) &&
		!(g_ulDontShowFlags & s_anWarnBits[ m_nSystem - 1 ]) )
	{
		CWarningDlg	dlgWarning;
		char szMessage[ LOADSTRING_STRING_SIZE ];
		LoadString( NULL, IDS_WARNING_ROM_CORRUPT, szMessage, LOADSTRING_STRING_SIZE );
		dlgWarning.csWarnText = szMessage;
		dlgWarning.iWarnBit   = s_anWarnBits[ m_nSystem - 1 ];
		nResult = dlgWarning.DoModal();
	}

	if( nResult == IDOK )
	{
		strcpy( s_acNewRomNames[ m_nSystem - 1 ], szBuffer );
		if( m_nSystem > 2 && m_nSystem < MACHINE_TYPES - 1 )
		{
			strcpy( s_acNewRomNames[ 2 ], szBuffer );
			strcpy( s_acNewRomNames[ 3 ], szBuffer );
			strcpy( s_acNewRomNames[ 4 ], szBuffer );
			strcpy( s_acNewRomNames[ 5 ], szBuffer );
		}
	}
	SetDlgState();

	return nResult;
} /* #OF# CHardwareDlg::OnKillfocusRomEdit */

/*========================================================
Method   : CHardwareDlg::OnSelchangeMachineSelector
=========================================================*/
/* #FN#
   Allows to select ROM image using FileDialog window */
void
/* #AS#
   Nothing */
CHardwareDlg::
OnRomfile()
{
	char        szBuffer[ MAX_PATH ];
	CString	    strName;
	int	        nResult;
	CFileDialog	fdlgRomImage( TRUE, NULL, NULL, OFN_EXPLORER | OFN_FILEMUSTEXIST, s_szFilter, this );

	fdlgRomImage.m_ofn.lpstrTitle = "Select the ROM for this machine type";
	nResult = fdlgRomImage.DoModal();
	if( nResult == IDOK )
	{
		strName = fdlgRomImage.GetPathName();
		GetDlgItemText( IDC_ROM_EDIT, szBuffer, MAX_PATH );
		SetDlgItemText( IDC_ROM_EDIT, strName.GetBuffer( 0 ) );
		nResult = OnKillfocusRomEdit();
		if( nResult != IDOK )
			SetDlgItemText( IDC_ROM_EDIT, szBuffer );
	}
} /* #OF# CHardwareDlg::OnRomfile */

/*========================================================
Method   : CHardwareDlg::OnDeltaposRefreshSpin
=========================================================*/
/* #FN#
	Sets the state of object regarding to Refresh spin */
void
/* #AS#
   Nothing */
CHardwareDlg::
OnDeltaposRefreshSpin( NMHDR   *pNMHDR, /* #IN#  */
					   LRESULT *pResult /* #OUT# */ )
{
	char       szDigits[ 64 ];
	NM_UPDOWN *pNMUpDown = (NM_UPDOWN*)pNMHDR;

	m_nRefreshRate -= pNMUpDown->iDelta;
	if( m_nRefreshRate < 1 )
		m_nRefreshRate = 1;
	if( m_nRefreshRate > 99 )
		m_nRefreshRate = 99;

	itoa( m_nRefreshRate, szDigits, 10 );
	SetDlgItemText( IDC_REFRESH_RATE, szDigits );

	test_val = 0;
	*pResult = 0;
} /* #OF# CHardwareDlg::OnDeltaposRefreshSpin */

/*========================================================
Method   : CHardwareDlg::OnKillfocusRefreshRate
=========================================================*/
/* #FN#
	Sets the state of object regarding to Refresh spin */
void
/* #AS#
   Nothing */
CHardwareDlg::
OnKillfocusRefreshRate() 
{
	char szDigits[ 64 ];
	int  iTemp;

	GetDlgItemText( IDC_REFRESH_RATE, szDigits, 64 );
	iTemp = atoi( szDigits );

	if( iTemp < 1 || iTemp > 99 )
	{
		itoa( m_nRefreshRate, szDigits, 10 );
		SetDlgItemText( IDC_REFRESH_RATE, szDigits );
	}
	else
		m_nRefreshRate = iTemp;
} /* #OF# CHardwareDlg::OnKillfocusRefreshRate */

/*========================================================
Method   : CHardwareDlg::OnNtsc
=========================================================*/
/* #FN#
	Sets the state of object regarding to NTSC radio */
void
/* #AS#
   Nothing */
CHardwareDlg::
OnNtsc() 
{
	m_nDefaultTVMode = 2;
} /* #OF# CHardwareDlg::OnNtsc */

/*========================================================
Method   : CHardwareDlg::OnPal
=========================================================*/
/* #FN#
	Sets the state of object regarding to PAL radio */
void
/* #AS#
   Nothing */
CHardwareDlg::
OnPal() 
{
	m_nDefaultTVMode = 1;
} /* #OF# CHardwareDlg::OnPal */

/*========================================================
Method   : CHardwareDlg::OnEnableRom
=========================================================*/
/* #FN#
	Sets the state of object regarding to Enable ROM Patches checkbox */
void
/* #AS#
   Nothing */
CHardwareDlg::
OnEnableRom() 
{
	CButton *pButton = (CButton*)GetDlgItem( IDC_ENABLE_ROM );
	m_nEnableRomPatches = pButton->GetCheck();

	SetDlgState();
} /* #OF# CHardwareDlg::OnEnableRom */

/*========================================================
Method   : CHardwareDlg::OnNewBinload
=========================================================*/
/* #FN#
	Sets the state of object regarding to Use New Binload checkbox */
void
/* #AS#
   Nothing */
CHardwareDlg::
OnNewBinload() 
{
	CButton *pButton = (CButton*)GetDlgItem( IDC_NEW_BINLOAD );
	m_nUseNewBinload = pButton->GetCheck();
} /* #OF# CHardwareDlg::OnNewBinload */

/*========================================================
Method   : CHardwareDlg::OnFullSpeed
=========================================================*/
/* #FN#
	Sets the state of object regarding to Full Speed checkbox */
void
/* #AS#
   Nothing */
CHardwareDlg::
OnFullSpeed() 
{
	CButton *pButton = (CButton*)GetDlgItem( IDC_FULL_SPEED );
	ASSERT(pButton);
	if( pButton->GetCheck() )
		m_ulMiscStates |= ATARI_FULL_SPEED;
	else
		m_ulMiscStates &= ~ATARI_FULL_SPEED;
} /* #OF# CHardwareDlg::OnFullSpeed */

/*========================================================
Method   : CHardwareDlg::OnEnableSio
=========================================================*/
/* #FN#
	Sets the state of object regarding to Enable SIO Patch checkbox */
void
/* #AS#
   Nothing */
CHardwareDlg::
OnEnableSio() 
{
	CButton *pButton = (CButton*)GetDlgItem( IDC_ENABLE_SIO );
	m_nEnableSIOPatch = pButton->GetCheck();
} /* #OF# CHardwareDlg::OnEnableSio */

/*========================================================
Method   : CHardwareDlg::OnC000Ram
=========================================================*/
/* #FN#
	Sets the state of object regarding to Enable C000 RAM checkbox */
void
/* #AS#
   Nothing */
CHardwareDlg::
OnC000Ram() 
{
	CButton *pButton = (CButton*)GetDlgItem( IDC_C000_RAM );
	m_nEnableC000Ram = pButton->GetCheck();
} /* #OF# CHardwareDlg::OnC000Ram */

/*========================================================
Method   : CHardwareDlg::OnOK
=========================================================*/
/* #FN#
   Handles the pressing OK button event */
void
/* #AS#
   Nothing */
CHardwareDlg::
OnOK() 
{
	BOOL bReboot = FALSE;
	BOOL bSaveOS = FALSE;

	CDialog::OnOK();

	if( g_ulAtariState & ATARI_LOAD_FAILED )
		bReboot = TRUE;

	if( m_nGlobalArtifMode != global_artif_mode )
	{
		global_artif_mode = m_nGlobalArtifMode;
		art_main_init( 0, global_artif_mode );
//		bReboot = TRUE;
		WriteRegDWORD( NULL, REG_ARTIF_MODE, global_artif_mode );
	}

	if( m_nEnableC000Ram != enable_c000_ram )
	{
		enable_c000_ram = m_nEnableC000Ram;
		bReboot = TRUE;
		WriteRegDWORD( NULL, REG_ENABLE_C000_RAM, enable_c000_ram );
	}

	if( m_nEnableRomPatches != enable_rom_patches )
	{
		enable_rom_patches = m_nEnableRomPatches;
		bReboot = TRUE;
		WriteRegDWORD( NULL, REG_ENABLE_ROM_PATCH, enable_rom_patches );
	}

	if( m_nEnableSIOPatch != enable_sio_patch )
	{
		enable_sio_patch = m_nEnableSIOPatch;
		if( enable_sio_patch )
			SetSIOEsc();
		else
			RestoreSIO();
		WriteRegDWORD( NULL, REG_ENABLE_SIO_PATCH, enable_sio_patch );
	}

	if( m_nUseNewBinload != use_new_binload )
	{
		use_new_binload = m_nUseNewBinload;
		bReboot = TRUE;
		WriteRegDWORD( NULL, REG_USE_NEW_BINLOAD, use_new_binload );
	}

	for( int i = 0 ; i < MACHINE_TYPES; i++ )
	{
		if( _stricmp( s_apszRomNames[ i ], s_acNewRomNames[ i ] ) )
			bSaveOS = TRUE;
		strcpy( s_apszRomNames[ i ], s_acNewRomNames[ i ] );
	}

	if( bSaveOS )
	{
		WriteRegString( NULL, REG_OSA_ROM,  atari_osa_filename  );
		WriteRegString( NULL, REG_OSB_ROM,  atari_osb_filename  );
		WriteRegString( NULL, REG_XLXE_ROM, atari_xlxe_filename );
		WriteRegString( NULL, REG_5200_ROM, atari_5200_filename );
	}

	if( m_nSystem != m_nOrigSystem )
	{
		switch( m_nSystem )
		{
			case 1:
				default_system = 1;		// OS-A
				break;
			case 2:
				default_system = 2;		// OS-B
				break;
			case 3:
				default_system = 3;		// 800 XL
				break;
			case 4:
				default_system = 4;		// 130 XE
				break;
			case 5:
				default_system = 5;
				Ram256 = 2;
				m_ulMiscStates &= ~ATARI_RAMBO_MODE;
				break;
			case 6:
				default_system = 5;
				Ram256 = 1;
				m_ulMiscStates |= ATARI_RAMBO_MODE;
				break;
			case 7:
				default_system = 6;
				break;
			default:
				default_system = 2;		// OS-B
				break;
		}
		bReboot = TRUE;
		WriteRegDWORD( NULL, REG_DEFAULT_SYSTEM, default_system );
	}

	if( m_nRefreshRate != refresh_rate )
	{
		refresh_rate = m_nRefreshRate;
		WriteRegDWORD( NULL, REG_REFRESH_RATE, refresh_rate );
	}

	if( m_nDefaultTVMode != default_tv_mode )
	{
		default_tv_mode = m_nDefaultTVMode;
		bReboot = TRUE;
		WriteRegDWORD( NULL, REG_DEFAULT_TV_MODE, default_tv_mode );
	}

	if( m_ulMiscStates != g_ulMiscStates )
	{
		g_ulMiscStates = m_ulMiscStates;
		WriteRegDWORD( NULL, REG_MISC_STATES, g_ulMiscStates );
	}

	/* If the Atari load failed, we just reset g_hWnd to NULL and the
	   Atari will be rebooted when the screen is refreshed in OnDraw */
	if( bReboot )
	{
		Atari_ReInit();
	}
} /* #OF# CHardwareDlg::OnOK */
